--#!/usr/bin/env python
--[[Generate a perfect maze, that is a maze with
    * all included points connected (no isolated passages) and
    * no loops (only one path between any two points).
The size and branching factor of the maze can be adjusted.

Coding by d.factorial [at] gmail.com.
]]


local function generation(xwide,yhigh)
	local random = {}

	function random.shuffle(t)
		local c = #t
		for i = 1, c do
			table.insert(t,table.remove(t,math.random(1,c)))
		end
	end

	local function array_concat(...)
		local t = {}
		for n = 1,select("#",...) do
			local arg = select(n,...)
			if type(arg)=="table" then
				for _,v in ipairs(arg) do
					t[#t+1] = v
				end
			else
				t[#t+1] = arg
			end
		end
		return t
	end

	--the dimensions of the maze

	--#the grid of the maze
	--#each cell of the maze is one of the following:
	--    # 1  = '#' is wall
	--    # 0  = '.' is empty space
	--    # -2 = ',' is exposed but undetermined
	--    # -1 = '?' is unexposed and undetermined
	field = {}

	for y = 1, yhigh do
		row = {}
		for x = 1, xwide do
			row[x] = -1
		end
		field[y] = row
	end

	--#list of coordinates of exposed but undetermined cells.
	frontier = {}

	local function carve(y, x)
		--[['''Make the cell at y,x a space.

		Update the fronteer and field accordingly.
		Note: this does not remove the current cell from frontier, it only adds new cells.
		''']]
		extra = {}
		field[y][x] = 0
		if x-1 > 0 then
			if field[y][x-1] == -1 then
				field[y][x-1] = -2
				table.insert(extra,{y,x-1})
			end
		end
		if x+1 < xwide+1 then
			if field[y][x+1] == -1 then
				field[y][x+1] = -2
				table.insert(extra,{y,x+1})
			end
		end
		if y-1 > 0 then
			if field[y-1][x] == -1 then
				field[y-1][x] = -2
				table.insert(extra,{y-1,x})
			end
		end
		if y+1 < yhigh+1 then
			if field[y+1][x] == -1 then
				field[y+1][x] = -2
				table.insert(extra,{y+1,x})
			end
		end

		random.shuffle(extra)

		frontier = array_concat(frontier,extra)
	end

	local function harden(y, x)
		--[['''Make the cell at y,x a wall.
		''']]
		field[y][x] = 1
	end


	local function check(y, x, nodiagonals)
		nodiagonals = nodiagonals or true
		--[['''Test the cell at y,x: can this cell become a space?

		true indicates it should become a space,
		false indicates it should become a wall.
		''']]

		edgestate = 0
		if x-1 > 0 then
			if field[y][x-1] == 0 then
				edgestate = edgestate + 1
			end
		end
		if x+1 < xwide+1 then
			if field[y][x+1] == 0 then
				edgestate = edgestate + 2
			end
		end
		if y-1 > 0 then
			if field[y-1][x] == 0 then
				edgestate = edgestate + 4
			end
		end
		if y+1 < yhigh+1 then
			if field[y+1][x] == 0 then
				edgestate = edgestate + 8
			end
		end

		if nodiagonals then
			--#if this would make a diagonal connecition, forbid it
				--#the following steps make the test a bit more complicated and are not necessary,
				--#but without them the mazes don't look as good
			if edgestate == 1 then
				if x+1 < xwide+1 then
					if y-1 > 0 then
						if field[y-1][x+1] == 0 then
							return false
						end
					end
					if y+1 < yhigh+1 then
						if field[y+1][x+1] == 0 then
							return false
						end
					end
				end
				return true
			elseif edgestate == 2 then
				if x-1 > 0 then
					if y-1 > 0 then
						if field[y-1][x-1] == 0 then
							return false
						end
					end
					if y+1 < yhigh+1 then
						if field[y+1][x-1] == 0 then
							return false
						end
					end
				end
				return true
			elseif edgestate == 4 then
				if y+1 < yhigh+1 then
					if x-1 > 0 then
						if field[y+1][x-1] == 0 then
							return false
						end
					end
					if x+1 < xwide+1 then
						if field[y+1][x+1] == 0 then
							return false
						end
					end
				end
				return true
			elseif edgestate == 8 then
				if y-1 > 0 then
					if x-1 > 0 then
						if field[y-1][x-1] == 0 then
							return false
						end
					end
					if x+1 < xwide+1 then
						if field[y-1][x+1] == 0 then
							return false
						end
					end
				end
				return true
			end
			return false
		else
			--#diagonal walls are permitted
			--[[if  [1,2,4,8].count(edgestate) then
				return true
			end
			return false]]
		end
	end


	--#choose a original point at random and carve it out.
	local xchoice = math.ceil(math.random()*xwide) --math.random(field.xn, field.xp)
	local ychoice = math.ceil(math.random()*yhigh) --math.random(field.yn, field.yp)
	carve(ychoice,xchoice)

	--#parameter branchrate:
		--#zero is unbiased, positive will make branches more frequent, negative will cause long passages
		--#this controls the position in the list chosen: positive makes the start of the list more likely,
		--#negative makes the end of the list more likely

		--#large negative values make the original point obvious

		--#try values between -10, 10
	local branchrate = 0

	--from math import e

	while #frontier ~= 0 do
		--#select a random edge
		pos = math.random()
		--pos = pos**(e**-branchrate)
		local c = math.floor(pos*(#frontier-1) + 1)
		choice = frontier[c]
		if check(choice[1],choice[2]) then
			carve(choice[1],choice[2])
		else
			harden(choice[1],choice[2])
		end
		table.remove(frontier,c)
	end

	--#set unexposed cells to be walls
	for y = 1, yhigh do
		for x = 1, xwide do
			if field[y][x] == -1 then
				field[y][x] = 1
			end
		end
	end

	--file = love.filesystem.newFile( "MAP.txt" )
	--file:open("w")

	--#print the maze
	--[[for y = 1, yhigh do
		s = ''
		for x = 1, xwide do
			s = s .. field[y][x]
		end
		file:write(s .. "NEWLINE")
		print(s)
	end]]

	--file:close()
	return field
end

return generation
